/*
 * dv_cminterf.c -- interface with datahub.
 *         Creates a datahub server and establishes the input mechanism -
 *         file or socket.  Then accepts events as they arrive.  Called with
 *         one of the following options and arguments:
 *                    -usock_in unix_socket_name
 *                    -sock_in  unix_socket_name
 *                    -isock_in inet_socket_name inet_port_number
 *                    -file_in  file_name
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <sys/un.h>
#include <assert.h>
#include <poll.h>

#include "dv_cminterf.h"
#include "comm_group.h"

typedef enum _InOutMethod {
    BY_NONE,
    BY_FILE,
    BY_SOCKET_INET,
    BY_SOCKET_UNIX,
} InOutMethod, *InOutMethodPtr;

extern int errno;
extern int printf(const char *, ...);
extern int fprintf(FILE *, const char *, ...);
extern char *strdup();
extern int atoi();
extern int close();
extern char* malloc();
extern char* realloc();
extern void free();
extern int getpid();
extern int unlink();
extern int poll();

/* global variables */
DM_Type_thread_init  *thread_init_data = NULL;
char            *event_buffer;
int             event_buffer_size = 1024;
DExchange		my_de;
DEPort  in_dep;

/* locally defined variables */
static InOutMethod  input_method, output_method;
static IOFormat listof_event_fmts[NUMBER_OF_EVENTS];
static EventHandlerProc listof_event_handlers[NUMBER_OF_EVENTS];
static EventHandlerProc default_event_handler = NULL;


/* command event formats */

static IOField ur_simple_command_flds[] = {
    {"type", "integer", sizeof(short), IOOffset(URSimpleCommandPtr,type)},
    {"proc", "integer", sizeof(short), IOOffset(URSimpleCommandPtr, proc)},
    {"sensor", "integer", sizeof(short), IOOffset(URSimpleCommandPtr,sensor)},
    {"int_field", "integer", sizeof(int),
     IOOffset(URSimpleCommandPtr, int_field)},
    {"aux_int", "integer", sizeof(int),
     IOOffset(URSimpleCommandPtr, aux_int)},
    {NULL, NULL, 0, 0}
};

static IOField ur_float_command_flds[] = {
    {"type", "integer", sizeof(short), IOOffset(URFloatCommandPtr,type)},
    {"proc", "integer", sizeof(short), IOOffset(URFloatCommandPtr, proc)},
    {"sensor", "integer", sizeof(short), IOOffset(URFloatCommandPtr,sensor)},
    {"float_field", "float", sizeof(double),
     IOOffset(URFloatCommandPtr, float_field)},
    {"aux_float", "float", sizeof(double),
     IOOffset(URFloatCommandPtr, aux_float)},
    {NULL, NULL, 0, 0}
};

static IOField ur_named_command_flds[] = {
    {"type", "integer", sizeof(short), IOOffset(URNamedCommandPtr, type)},
    {"proc", "integer", sizeof(short), IOOffset(URNamedCommandPtr, proc)},
    {"sensor", "integer", sizeof(short), IOOffset(URNamedCommandPtr,sensor)},
    {"file_name", "string", sizeof(char*),
     IOOffset(URNamedCommandPtr, file_name)},
    {"aux_name", "string", sizeof(char*),
     IOOffset(URNamedCommandPtr, aux_name)},
    {NULL, NULL, 0, 0}
};


/* trace event formats */

static IOField DM_Format_thread_init[] = {
    {"type", "integer", sizeof(int), IOOffset(SimpleTraceDataRecordPtr, type)},
    {"perturbation", "integer", sizeof(int),
       IOOffset(SimpleTraceDataRecordPtr, perturbation)},
    {"timestamp", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, timestamp)},
    {"thread", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, thread)},
    {"name", "string", sizeof(char *),
       IOOffset(DM_Type_thread_initPtr, name)},
    {"num_processors", "integer", sizeof(int),
       IOOffset(DM_Type_thread_initPtr, num_processors)},
    {"num_lms", "integer", sizeof(int),
       IOOffset(DM_Type_thread_initPtr, num_lms)},
    {"on_processor", "integer", sizeof(int),
       IOOffset(DM_Type_thread_initPtr, on_processor)},
    {"timestamp_resolution", "unsigned integer", sizeof(unsigned long),
       IOOffset(DM_Type_thread_initPtr, timestamp_resolution)},
    {NULL, NULL, 0, 0}
};

static IOField DM_Format_thread_fork[] = {
    {"type", "integer", sizeof(int), IOOffset(SimpleTraceDataRecordPtr, type)},
    {"perturbation", "integer", sizeof(int),
       IOOffset(SimpleTraceDataRecordPtr, perturbation)},
    {"timestamp", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, timestamp)},
    {"thread", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, thread)},
    {"name", "string", sizeof(char *),
       IOOffset(DM_Type_thread_forkPtr, name)},
    {"child_thread", "unsigned integer", sizeof(unsigned long),
       IOOffset(DM_Type_thread_forkPtr, child_thread)},
    {"sequence_number", "integer", sizeof(int),
       IOOffset(DM_Type_thread_forkPtr, sequence_number)},
    {"on_processor", "integer", sizeof(int),
       IOOffset(DM_Type_thread_forkPtr, on_processor)},
    {NULL, NULL, 0, 0}
};

static IOField DM_Format_thread_begin_join[] = {
    {"type", "integer", sizeof(int), IOOffset(SimpleTraceDataRecordPtr, type)},
    {"perturbation", "integer", sizeof(int),
       IOOffset(SimpleTraceDataRecordPtr, perturbation)},
    {"timestamp", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, timestamp)},
    {"thread", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, thread)},
    {"join_to", "unsigned integer", sizeof(unsigned long),
       IOOffset(DM_Type_thread_begin_joinPtr, join_to)},
    {NULL, NULL, 0, 0}
};

static IOField DM_Format_thread_end_join[] = {
    {"type", "integer", sizeof(int), IOOffset(SimpleTraceDataRecordPtr, type)},
    {"perturbation", "integer", sizeof(int),
       IOOffset(SimpleTraceDataRecordPtr, perturbation)},
    {"timestamp", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, timestamp)},
    {"thread", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, thread)},
    {"join_to", "unsigned integer", sizeof(unsigned long),
       IOOffset(DM_Type_thread_end_joinPtr, join_to)},
    {NULL, NULL, 0, 0}
};

static IOField DM_Format_thread_exit[] = {
    {"type", "integer", sizeof(int), IOOffset(SimpleTraceDataRecordPtr, type)},
    {"perturbation", "integer", sizeof(int),
       IOOffset(SimpleTraceDataRecordPtr, perturbation)},
    {"timestamp", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, timestamp)},
    {"thread", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, thread)},
    {NULL, NULL, 0, 0}
};

static IOField DM_Format_thread_detach[] = {
    {"type", "integer", sizeof(int), IOOffset(SimpleTraceDataRecordPtr, type)},
    {"perturbation", "integer", sizeof(int),
       IOOffset(SimpleTraceDataRecordPtr, perturbation)},
    {"timestamp", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, timestamp)},
    {"thread", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, thread)},
    {"detach_from", "unsigned integer", sizeof(unsigned long),
       IOOffset(DM_Type_thread_detachPtr, detach_from)},
    {NULL, NULL, 0, 0}
};

static IOField DM_Format_thread_set_name[] = {
    {"type", "integer", sizeof(int), IOOffset(SimpleTraceDataRecordPtr, type)},
    {"perturbation", "integer", sizeof(int),
       IOOffset(SimpleTraceDataRecordPtr, perturbation)},
    {"timestamp", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, timestamp)},
    {"thread", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, thread)},
    {"the_thread", "unsigned integer", sizeof(unsigned long),
       IOOffset(DM_Type_thread_set_namePtr, the_thread)},
    {"name", "string", sizeof(char *),
       IOOffset(DM_Type_thread_set_namePtr, name)},
    {NULL, NULL, 0, 0}
};

static IOField DM_Format_mutex_begin_lock[] = {
    {"type", "integer", sizeof(int), IOOffset(SimpleTraceDataRecordPtr, type)},
    {"perturbation", "integer", sizeof(int),
       IOOffset(SimpleTraceDataRecordPtr, perturbation)},
    {"timestamp", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, timestamp)},
    {"thread", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, thread)},
    {"mutex", "unsigned integer", sizeof(unsigned long),
       IOOffset(DM_Type_mutex_begin_lockPtr, mutex)},
    {"sequence_number", "integer", sizeof(int),
       IOOffset(DM_Type_mutex_begin_lockPtr, sequence_number)},
    {NULL, NULL, 0, 0}
};

static IOField DM_Format_mutex_end_lock[] = {
    {"type", "integer", sizeof(int), IOOffset(SimpleTraceDataRecordPtr, type)},
    {"perturbation", "integer", sizeof(int),
       IOOffset(SimpleTraceDataRecordPtr, perturbation)},
    {"timestamp", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, timestamp)},
    {"thread", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, thread)},
    {"mutex", "unsigned integer", sizeof(unsigned long),
       IOOffset(DM_Type_mutex_end_lockPtr, mutex)},
    {"sequence_number", "integer", sizeof(int),
       IOOffset(DM_Type_mutex_end_lockPtr, sequence_number)},
    {NULL, NULL, 0, 0}
};

static IOField DM_Format_mutex_set_name[] = {
    {"type", "integer", sizeof(int), IOOffset(SimpleTraceDataRecordPtr, type)},
    {"perturbation", "integer", sizeof(int),
       IOOffset(SimpleTraceDataRecordPtr, perturbation)},
    {"timestamp", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, timestamp)},
    {"thread", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, thread)},
    {"mutex", "unsigned integer", sizeof(unsigned long),
       IOOffset(DM_Type_mutex_set_namePtr, mutex)},
    {"name", "string", sizeof(char *),
       IOOffset(DM_Type_mutex_set_namePtr, name)},
    {NULL, NULL, 0, 0}
};

static IOField DM_Format_condition_begin_wait[] = {
    {"type", "integer", sizeof(int), IOOffset(SimpleTraceDataRecordPtr, type)},
    {"perturbation", "integer", sizeof(int),
       IOOffset(SimpleTraceDataRecordPtr, perturbation)},
    {"timestamp", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, timestamp)},
    {"thread", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, thread)},
    {"condition", "unsigned integer", sizeof(unsigned long),
       IOOffset(DM_Type_condition_begin_waitPtr, condition)},
    {"sequence_number", "integer", sizeof(int),
       IOOffset(DM_Type_condition_begin_waitPtr, sequence_number)},
    {"mutex", "unsigned integer", sizeof(unsigned long),
       IOOffset(DM_Type_condition_begin_waitPtr, mutex)},
    {NULL, NULL, 0, 0}
  };

static IOField DM_Format_condition_end_wait[] = {
    {"type", "integer", sizeof(int), IOOffset(SimpleTraceDataRecordPtr, type)},
    {"perturbation", "integer", sizeof(int),
       IOOffset(SimpleTraceDataRecordPtr, perturbation)},
    {"timestamp", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, timestamp)},
    {"thread", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, thread)},
    {"condition", "unsigned integer", sizeof(unsigned long),
       IOOffset(DM_Type_condition_end_waitPtr, condition)},
    {"sequence_number", "integer", sizeof(int),
       IOOffset(DM_Type_condition_end_waitPtr, sequence_number)},
    {"mutex", "unsigned integer", sizeof(unsigned long),
       IOOffset(DM_Type_condition_end_waitPtr, mutex)},
    {NULL, NULL, 0, 0}
};

static IOField DM_Format_condition_set_name[] = {
    {"type", "integer", sizeof(int), IOOffset(SimpleTraceDataRecordPtr, type)},
    {"perturbation", "integer", sizeof(int),
       IOOffset(SimpleTraceDataRecordPtr, perturbation)},
    {"timestamp", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, timestamp)},
    {"thread", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, thread)},
    {"condition", "unsigned integer", sizeof(unsigned long),
       IOOffset(DM_Type_condition_set_namePtr, condition)},
    {"name", "string", sizeof(char *),
       IOOffset(DM_Type_condition_set_namePtr, name)},
    {NULL, NULL, 0, 0}
};

static IOField DM_Format_thread_off_runq[] = {
    {"type", "integer", sizeof(int), IOOffset(SimpleTraceDataRecordPtr, type)},
    {"perturbation", "integer", sizeof(int),
       IOOffset(SimpleTraceDataRecordPtr, perturbation)},
    {"timestamp", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, timestamp)},
    {"thread", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, thread)},
    {"thread_off_runq", "unsigned integer", sizeof(unsigned long),
       IOOffset(DM_Type_thread_off_runqPtr, thread_off_runq)},
    {NULL, NULL, 0, 0}
};

static IOField DM_Format_monitor_exit[] = {
    {"type", "integer", sizeof(int), IOOffset(SimpleTraceDataRecordPtr, type)},
    {"perturbation", "integer", sizeof(int),
       IOOffset(SimpleTraceDataRecordPtr, perturbation)},
    {"timestamp", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, timestamp)},
    {"thread", "unsigned integer", sizeof(unsigned long),
       IOOffset(SimpleTraceDataRecordPtr, thread)},
    {"lm_number", "integer", sizeof(int),
       IOOffset(DM_Type_monitor_exitPtr, lm_number)},
    {NULL, NULL, 0, 0}
};



/*** array of default event fields ***/

static IOField *listof_event_fields[NUMBER_OF_EVENTS] = {
    ur_simple_command_flds,
    ur_float_command_flds,
    ur_named_command_flds,
    DM_Format_thread_init,          	/* thread_init */
    DM_Format_thread_fork,		/* thread_fork */
    DM_Format_thread_exit,    		/* thread_exit */
    DM_Format_thread_begin_join,	/* thread_begin_join */
    DM_Format_thread_end_join,		/* thread_end_join */
    DM_Format_thread_detach,   		/* thread_detach */
    DM_Format_thread_set_name,		/* thread_set_name */
    DM_Format_mutex_begin_lock,		/* mutex_begin_lock */
    DM_Format_mutex_end_lock,		/* mutex_end_lock */
    DM_Format_mutex_set_name,		/* mutex_set_name */
    DM_Format_condition_begin_wait,	/* condition_begin_wait */
    DM_Format_condition_end_wait,	/* condition_end_wait */
    DM_Format_condition_set_name,	/* condition_set_name */
    DM_Format_thread_off_runq,		/* off_readyq */
    DM_Format_monitor_exit,		/* end of monitor thread */
};


/*** sizes of event types ***/
static int sizeof_event_types[NUMBER_OF_EVENTS] = {
    sizeof(URSimpleCommand),
    sizeof(URFloatCommand),
    sizeof(URNamedCommand),
    sizeof(DM_Type_thread_init),        	/* thread_init */
    sizeof(DM_Type_thread_fork),		/* thread_fork */
    sizeof(DM_Type_thread_exit),    		/* thread_exit */
    sizeof(DM_Type_thread_begin_join),		/* thread_begin_join */
    sizeof(DM_Type_thread_end_join),		/* thread_end_join */
    sizeof(DM_Type_thread_detach),   		/* thread_detach */
    sizeof(DM_Type_thread_set_name),		/* thread_set_name */
    sizeof(DM_Type_mutex_begin_lock),		/* mutex_begin_lock */
    sizeof(DM_Type_mutex_end_lock),		/* mutex_end_lock */
    sizeof(DM_Type_mutex_set_name),		/* mutex_set_name */
    sizeof(DM_Type_condition_begin_wait),	/* condition_begin_wait */
    sizeof(DM_Type_condition_end_wait),		/* condition_end_wait */
    sizeof(DM_Type_condition_set_name),		/* condition_set_name */
    sizeof(DM_Type_thread_off_runq),		/* off_readyq */
    sizeof(DM_Type_monitor_exit),		/* end of monitor thread */
};

/*** array of default event names ***/

static char *listof_event_names[NUMBER_OF_EVENTS] = {
    "simple user request",
    "float user request",
    "named user request",
    "cthread_init",
    "cthread_fork",
    "cthread_exit",
    "cthread_begin_join",
    "cthread_end_join",
    "cthread_detach",
    "cthread_set_name",
    "mutex_begin_lock",
    "mutex_end_lock",
    "mutex_set_name",
    "condition_begin_wait",
    "condition_end_wait",
    "condition_set_name",
    "thread_off_runq",
    "monitor_exit",
};


static char		*app_list_default = "*";
static char		*type_list_default = "reorder,*";
extern int debug_flag;

int
CMonInterfInit(iargc, iargv)
int  *iargc;
char **iargv;
{
    int  i, num_args, newargc = 0;
    char *in_name, *program_name;
    int  in_port_num = -1;

    program_name = iargv[0]+strlen(iargv[0])-1;
    while (*program_name != '/' && program_name != iargv[0]) program_name--;
    if (*program_name == '/') program_name++;
    i = 1;
    input_method = output_method = BY_NONE;
    in_name = NULL;
    num_args = *iargc;
    while (i < num_args) {
	if (!strcmp(iargv[i], "-debug")) {
	    debug_flag++;
	    iargv[i] = NULL;
	} else if (!strncmp(iargv[i], "-usock_in", 6) ||
	    !strncmp(iargv[i], "-sock_in", 5)) {
	    if (++i + 1 > num_args) {
		fprintf(stderr, "too few arguments to %s", iargv[i-1]);
		return -1;
	    }
	    input_method = BY_SOCKET_UNIX;
	    in_name = iargv[i];
	    iargv[i] = iargv[i-1] = NULL;
	    in_port_num = -1;
	} else if (!strncmp(iargv[i], "-isock_in", 6)) {
	    if (++i + 2 > num_args) {
		fprintf(stderr, "too few arguments to %s", iargv[i-1]);
		return -1;
	    }
	    input_method = BY_SOCKET_INET;
	    in_name = iargv[i++];
	    in_port_num = atoi(iargv[i]);
	    iargv[i] = iargv[i-1] = iargv[i-2] = NULL;
	} else if (!strncmp(iargv[i], "-file_in", 5)) {
	    if (++i == num_args) {
		fprintf(stderr, "too few arguments to %s", iargv[i-1]);
		return -1;
	    }
	    input_method = BY_FILE;
	    in_name = iargv[i];
	    iargv[i] = iargv[i-1] = NULL;
	}
	i++;
    }
    if (input_method == BY_NONE) {
	input_method = BY_SOCKET_INET;
    }
    iargv[newargc++] = iargv[0];
    for (i = 1; i < num_args; i++) {
	if (iargv[i] != NULL) iargv[newargc++] = iargv[i];
    }
    *iargc = newargc;

    /* create and init a datahub server */
    my_de = DExchange_create();

    /* open output socket */
    if (input_method == BY_SOCKET_INET || input_method == BY_SOCKET_UNIX) {
	if ((in_name != NULL ) && (in_port_num != -1)) {
	    in_dep = DExchange_initiate_conn(my_de, in_name, in_port_num, 
					     False);
	} else {
	    if (group_server_present()) {
		comm_group_return groups;
		groups = matching_comm_groups(app_list_default, type_list_default);
		in_dep = DExchange_initiate_first(my_de, groups, False);
	    } else {
		printf("Error:  No group server present and input unspecified\n");
		return -1;
	    }
	}
    } else if (input_method == BY_FILE) {
	in_dep = DExchange_accept_conn_file(my_de, in_name, "r");
    }
    if (in_dep == NULL) {
	fprintf(stderr, "failed to open input connection\n");
	return -1;
    }
    /* allocate space for event buffer */
    if (event_buffer == NULL)
	event_buffer = (char*) malloc(event_buffer_size);

    return 0;
}

extern void remove_input();
/* this procedure is called by the main program or XtEventHandler
   constantly.  It reads data, processes it, and draw new objects
   if needed.  If end of trace is detected, it returns TRUE, otherwise,
   it returns FALSE. */
int
ReadAndProcessInput()
{
    extern int stepping_flag;
    int  i, finished = False;
    char *comment;
    IOFormat new_format, next_format;
    char *format_name;

    if (stepping_flag > 0) stepping_flag--;
    switch (next_IOrecord_type(DEport_from_port(in_dep))) {
    case IOcomment:
	comment = read_comment_IOfile(DEport_from_port(in_dep));
	fprintf(stderr, "Comment: \"%s\"\n", comment);
	break;

    case IOformat:
	new_format = read_format_IOfile(DEport_from_port(in_dep));
	format_name = name_of_IOformat(new_format);
	for (i = 0; (i < NUMBER_OF_EVENTS) &&
		 (listof_event_names[i] == NULL ||
		  strcmp(format_name, listof_event_names[i]));
	     i++);
	if (i < NUMBER_OF_EVENTS) {
	    if (listof_event_fmts[i] != NULL)
		fprintf(stderr, "got format %s before\n", format_name);
	    listof_event_fmts[i] = new_format;
	    set_IOconversion(DEport_from_port(in_dep), format_name,
			     listof_event_fields[i],
			     sizeof_event_types[i]);
	} 
	break;

    case IOdata:
	next_format = next_IOrecord_format(DEport_from_port(in_dep));

	if (next_IOrecord_length(DEport_from_port(in_dep))
	    > event_buffer_size)
	    event_buffer = realloc(event_buffer);

	for (i = 0; (i < NUMBER_OF_EVENTS) &&
		 (listof_event_fmts[i] != next_format);
	     i++);
	if (i < NUMBER_OF_EVENTS) {
	    read_IOfile(DEport_from_port(in_dep), event_buffer);
	    if (listof_event_handlers[i] != NULL)
		listof_event_handlers[i](i, event_buffer);
	    else if (default_event_handler != NULL)
		default_event_handler(i, event_buffer);
	} else {
	    read_raw_IOfile(DEport_from_port(in_dep), NULL, 0, NULL);
	    if (debug_flag) {
		fprintf(stderr, "discarding a record of unknown format %s\n",
			name_of_IOformat(next_format));
	    }
	}
	break;

    case IOerror:
    case IOend:
	remove_input();
	fprintf(stderr, "got an ioerror or end ...\n");
	finished = True;
    }
    return finished;
}

EventHandlerProc
RegisterEventHandler(int ev_num, EventHandlerProc ev_proc)
{
    EventHandlerProc  old_proc = NULL;
    if (ev_num >= 0 && ev_num <	NUMBER_OF_EVENTS) {
	old_proc = listof_event_handlers[ev_num];
	listof_event_handlers[ev_num] = ev_proc;
    } else if (ev_num == -1) {
	/* if negative number, register it as default */
	old_proc = default_event_handler;
	default_event_handler = ev_proc;
    }
    return old_proc;
}
